@@ -491,12 +491,12 @@ static inline bool is_mesa_device(ID3D12Device *device)
     return false;
 }
 
-static inline bool is_mesa_device_lt(ID3D12Device *device, uint32_t major, uint32_t minor, uint32_t patch)
+static inline bool get_mesa_device_version(ID3D12Device *device, uint32_t *major, uint32_t *minor, uint32_t *patch)
 {
     return false;
 }
 
-static inline bool is_mesa_intel_device(ID3D12Device *device)
+static inline bool get_llvm_device_version(ID3D12Device *device, uint32_t *major, uint32_t *minor, uint32_t *patch)
 {
     return false;
 }
@@ -506,8 +506,7 @@ static inline bool is_llvmpipe_device(ID3D12Device *device)
     return false;
 }
 
-static inline bool is_llvmpipe_device_gte(ID3D12Device *device,
-        uint32_t major, uint32_t minor, uint32_t patch)
+static inline bool is_intel_device(ID3D12Device *device)
 {
     return false;
 }
@@ -543,6 +542,11 @@ static inline bool is_mvk_paravirtualized_device(ID3D12Device *device)
     return false;
 }
 
+static inline bool get_mvk_device_version(ID3D12Device *device, uint32_t *major, uint32_t *minor, uint32_t *patch)
+{
+    return false;
+}
+
 static inline bool is_mvk_device_lt(ID3D12Device *device, uint32_t major, uint32_t minor, uint32_t patch)
 {
     return false;
@@ -553,6 +557,11 @@ static inline bool is_macos_lt(unsigned int major, unsigned int minor, unsigned
     return false;
 }
 
+static inline bool is_cull_distance_supported(ID3D12Device *device)
+{
+    return true;
+}
+
 static inline bool is_depth_clip_enable_supported(ID3D12Device *device)
 {
     return true;
@@ -796,22 +805,24 @@ static inline bool is_mesa_device(ID3D12Device *device)
     return is_mesa_vulkan_driver(&properties);
 }
 
-static inline bool is_mesa_device_lt(ID3D12Device *device, uint32_t major, uint32_t minor, uint32_t patch)
-{
-    VkPhysicalDeviceDriverPropertiesKHR driver_properties;
-    VkPhysicalDeviceProperties device_properties;
-
-    get_driver_properties(device, &device_properties, &driver_properties);
-    return is_mesa_vulkan_driver(&driver_properties)
-            && !is_vulkan_driver_version_ge(&device_properties, &driver_properties, major, minor, patch);
-}
-
-static inline bool is_mesa_intel_device(ID3D12Device *device)
+static inline bool get_mesa_device_version(ID3D12Device *device, uint32_t *major, uint32_t *minor, uint32_t *patch)
 {
     VkPhysicalDeviceDriverPropertiesKHR properties;
+    VkPhysicalDeviceProperties device_properties;
 
-    get_driver_properties(device, NULL, &properties);
-    return properties.driverID == VK_DRIVER_ID_INTEL_OPEN_SOURCE_MESA_KHR;
+    get_driver_properties(device, &device_properties, &properties);
+
+    return get_mesa_driver_version(&device_properties, &properties, major, minor, patch);
+}
+
+static inline bool get_llvm_device_version(ID3D12Device *device, uint32_t *major, uint32_t *minor, uint32_t *patch)
+{
+    VkPhysicalDeviceDriverPropertiesKHR properties;
+    VkPhysicalDeviceProperties device_properties;
+
+    get_driver_properties(device, &device_properties, &properties);
+
+    return get_llvm_driver_version(&device_properties, major, minor, patch);
 }
 
 static inline bool is_llvmpipe_device(ID3D12Device *device)
@@ -822,15 +833,12 @@ static inline bool is_llvmpipe_device(ID3D12Device *device)
     return properties.driverID == VK_DRIVER_ID_MESA_LLVMPIPE;
 }
 
-static inline bool is_llvmpipe_device_gte(ID3D12Device *device,
-        uint32_t major, uint32_t minor, uint32_t patch)
+static inline bool is_intel_device(ID3D12Device *device)
 {
-    VkPhysicalDeviceDriverPropertiesKHR driver_properties;
-    VkPhysicalDeviceProperties device_properties;
+    VkPhysicalDeviceDriverPropertiesKHR properties;
 
-    get_driver_properties(device, &device_properties, &driver_properties);
-    return driver_properties.driverID == VK_DRIVER_ID_MESA_LLVMPIPE
-            && is_vulkan_driver_version_ge(&device_properties, &driver_properties, major, minor, patch);
+    get_driver_properties(device, NULL, &properties);
+    return properties.driverID == VK_DRIVER_ID_INTEL_OPEN_SOURCE_MESA;
 }
 
 static inline bool is_nvidia_device(ID3D12Device *device)
@@ -841,21 +849,19 @@ static inline bool is_nvidia_device(ID3D12Device *device)
     return properties.driverID == VK_DRIVER_ID_NVIDIA_PROPRIETARY_KHR;
 }
 
-/* NVIDIA uses a different bit pattern than standard Vulkan. */
-#define NV_MAKE_API_VERSION(major, minor, patch) \
-    ((((uint32_t)(major)) << 22u) | (((uint32_t)(minor)) << 14u) | (((uint32_t)(patch)) << 6u))
-
 static inline bool is_nvidia_device_lt(ID3D12Device *device,
         uint32_t major, uint32_t minor, uint32_t patch)
 {
     VkPhysicalDeviceDriverPropertiesKHR driver_properties;
+    uint32_t driver_major, driver_minor, driver_patch;
     VkPhysicalDeviceProperties device_properties;
 
     get_driver_properties(device, &device_properties, &driver_properties);
-    if (driver_properties.driverID != VK_DRIVER_ID_NVIDIA_PROPRIETARY_KHR)
+
+    if (!get_nvidia_driver_version(&device_properties, &driver_properties, &driver_major, &driver_minor, &driver_patch))
         return false;
 
-    return device_properties.driverVersion < NV_MAKE_API_VERSION(major, minor, patch);
+    return compare_versions(driver_major, driver_minor, driver_patch, major, minor, patch) < 0;
 }
 
 static inline bool is_qualcomm_device(ID3D12Device *device)
@@ -892,20 +898,24 @@ static inline bool is_mvk_paravirtualized_device(ID3D12Device *device)
             && strcmp(device_properties.deviceName, "Apple Paravirtual device") == 0;
 }
 
-/* MoltenVK uses a different pattern than standard Vulkan. */
-#define MVK_MAKE_API_VERSION(major, minor, patch) \
-    ((((uint32_t)(major)) * 10000) + (((uint32_t)(minor)) * 100) + (uint32_t)(patch))
+static inline bool get_mvk_device_version(ID3D12Device *device, uint32_t *major, uint32_t *minor, uint32_t *patch)
+{
+    VkPhysicalDeviceDriverPropertiesKHR properties;
+    VkPhysicalDeviceProperties device_properties;
+
+    get_driver_properties(device, &device_properties, &properties);
+
+    return get_mvk_driver_version(&device_properties, &properties, major, minor, patch);
+}
 
 static inline bool is_mvk_device_lt(ID3D12Device *device, uint32_t major, uint32_t minor, uint32_t patch)
 {
-    VkPhysicalDeviceDriverPropertiesKHR driver_properties;
-    VkPhysicalDeviceProperties device_properties;
+    uint32_t device_major, device_minor, device_patch;
 
-    get_driver_properties(device, &device_properties, &driver_properties);
-    if (driver_properties.driverID != VK_DRIVER_ID_MOLTENVK)
+    if (!get_mvk_device_version(device, &device_major, &device_minor, &device_patch))
         return false;
 
-    return device_properties.driverVersion < MVK_MAKE_API_VERSION(major, minor, patch);
+    return compare_versions(device_major, device_minor, device_patch, major, minor, patch) < 0;
 }
 
 #ifdef __APPLE__
@@ -940,6 +950,15 @@ static inline bool is_macos_lt(unsigned int major, unsigned int minor, unsigned
 
 #endif
 
+static inline bool is_cull_distance_supported(ID3D12Device *device)
+{
+    VkPhysicalDeviceFeatures features;
+
+    vkGetPhysicalDeviceFeatures(vkd3d_get_vk_physical_device(device), &features);
+
+    return features.shaderCullDistance;
+}
+
 static inline bool is_depth_clip_enable_supported(ID3D12Device *device)
 {
     VkPhysicalDevice vk_physical_device = vkd3d_get_vk_physical_device(device);
