@@ -258,6 +258,39 @@ static bool get_format_support(ID3D11Device *device, enum DXGI_FORMAT format)
     return ret;
 }
 
+static bool get_adapter_desc(ID3D11Device *device, DXGI_ADAPTER_DESC *desc)
+{
+    IDXGIDevice *dxgi_device;
+    IDXGIAdapter *adapter;
+    HRESULT hr;
+
+    memset(desc, 0, sizeof(*desc));
+
+    if (!vkd3d_test_platform_is_windows())
+        return false;
+
+    hr = ID3D11Device_QueryInterface(device, &IID_IDXGIDevice, (void **)&dxgi_device);
+    ok(hr == S_OK, "Failed to get DXGI device, hr %#lx.\n", hr);
+
+    hr = IDXGIDevice_GetAdapter(dxgi_device, &adapter);
+    ok(hr == S_OK, "Failed to get DXGI adapter, hr %#lx.\n", hr);
+
+    hr = IDXGIAdapter_GetDesc(adapter, desc);
+    ok(hr == S_OK, "Failed to get adapter desc, hr %#lx.\n", hr);
+
+    IDXGIAdapter_Release(adapter);
+    IDXGIDevice_Release(dxgi_device);
+
+    return true;
+}
+
+static bool is_nvidia_windows_device(ID3D11Device *device)
+{
+    DXGI_ADAPTER_DESC desc;
+
+    return get_adapter_desc(device, &desc) && desc.VendorId == 0x10de;
+}
+
 static BOOL init_test_context(struct d3d11_shader_runner *runner)
 {
     D3D11_FEATURE_DATA_D3D11_OPTIONS2 options2 = {0};
@@ -303,6 +336,7 @@ static BOOL init_test_context(struct d3d11_shader_runner *runner)
     runner->caps.minimum_shader_model = SHADER_MODEL_4_0;
     runner->caps.maximum_shader_model = SHADER_MODEL_5_0;
 
+    runner->caps.shader_caps[SHADER_CAP_CULL_DISTANCE] = true;
     hr = ID3D11Device_CheckFeatureSupport(runner->device, D3D11_FEATURE_DOUBLES,
             &doubles, sizeof(doubles));
     ok(hr == S_OK, "Failed to check double precision feature support, hr %#lx.\n", hr);
@@ -325,9 +359,14 @@ static BOOL init_test_context(struct d3d11_shader_runner *runner)
         runner->caps.format_caps[formats[i]] = get_format_support(runner->device, formats[i]);
     }
 
-    runner->caps.tag_count = 0;
+    runner->caps.tags[SHADER_RUNNER_TAG_D3D11].present = true;
+    runner->caps.tags[SHADER_RUNNER_TAG_WINDOWS].present = true;
+
     if (test_options.use_warp_device)
-        runner->caps.tags[runner->caps.tag_count++] = "warp";
+        runner->caps.tags[SHADER_RUNNER_TAG_WARP].present = true;
+
+    if (is_nvidia_windows_device(runner->device))
+        runner->caps.tags[SHADER_RUNNER_TAG_NVIDIA].present = true;
 
     rt_width = RENDER_TARGET_WIDTH;
     rt_height = RENDER_TARGET_HEIGHT;
@@ -423,7 +462,7 @@ static void init_subresource_data(D3D11_SUBRESOURCE_DATA *resource_data, const s
 
         for (unsigned int layer = 0; layer < params->desc.layer_count; ++layer)
         {
-            D3D11_SUBRESOURCE_DATA *subresource = &resource_data[level * params->desc.layer_count + layer];
+            D3D11_SUBRESOURCE_DATA *subresource = &resource_data[level + layer * params->desc.level_count];
             subresource->pSysMem = &params->data[buffer_offset];
             subresource->SysMemPitch = level_width * params->desc.texel_size;
             subresource->SysMemSlicePitch = level_height * subresource->SysMemPitch;
@@ -451,15 +490,11 @@ static void create_identity_view(ID3D11Device *device,
 static bool init_resource_2d(struct d3d11_shader_runner *runner, struct d3d11_resource *resource,
         const struct resource_params *params)
 {
-    D3D11_SUBRESOURCE_DATA resource_data[6];
     ID3D11Device *device = runner->device;
     D3D11_TEXTURE2D_DESC desc = {0};
     UINT quality_levels;
     HRESULT hr;
 
-    if (params->desc.level_count > ARRAY_SIZE(resource_data))
-        fatal_error("Level count %u is too high.\n", params->desc.level_count);
-
     if (params->desc.sample_count > 1)
     {
         if (params->desc.level_count > 1)
@@ -488,11 +523,15 @@ static bool init_resource_2d(struct d3d11_shader_runner *runner, struct d3d11_re
 
     if (params->data)
     {
+        D3D11_SUBRESOURCE_DATA *resource_data;
+
         if (params->desc.sample_count > 1)
             fatal_error("Cannot upload data to a multisampled texture.\n");
 
+        resource_data = calloc(params->desc.layer_count * params->desc.level_count, sizeof(*resource_data));
         init_subresource_data(resource_data, params);
         hr = ID3D11Device_CreateTexture2D(device, &desc, resource_data, &resource->texture);
+        free(resource_data);
     }
     else
     {
