@@ -83,14 +83,12 @@ static struct resource *d3d12_runner_create_resource(struct shader_runner *r, co
     struct d3d12_shader_runner *runner = d3d12_shader_runner(r);
     struct test_context *test_context = &runner->test_context;
     ID3D12Device *device = test_context->device;
-    D3D12_SUBRESOURCE_DATA resource_data[6] = {0};
     D3D12_RESOURCE_STATES initial_state, state;
+    D3D12_SUBRESOURCE_DATA *resource_data;
     struct d3d12_resource *resource;
     unsigned int buffer_offset = 0;
 
-    if (params->desc.level_count > ARRAY_SIZE(resource_data))
-        fatal_error("Level count %u is too high.\n", params->desc.level_count);
-
+    resource_data = calloc(params->desc.layer_count * params->desc.level_count, sizeof(*resource_data));
     resource = calloc(1, sizeof(*resource));
     init_resource(&resource->r, params);
 
@@ -102,7 +100,7 @@ static struct resource *d3d12_runner_create_resource(struct shader_runner *r, co
 
         for (unsigned int layer = 0; layer < params->desc.layer_count; ++layer)
         {
-            D3D12_SUBRESOURCE_DATA *subresource = &resource_data[level * params->desc.layer_count + layer];
+            D3D12_SUBRESOURCE_DATA *subresource = &resource_data[level + layer * params->desc.level_count];
             subresource->pData = &params->data[buffer_offset];
             subresource->RowPitch = level_width * params->desc.texel_size;
             subresource->SlicePitch = level_height * subresource->RowPitch;
@@ -145,10 +143,6 @@ static struct resource *d3d12_runner_create_resource(struct shader_runner *r, co
             break;
 
         case RESOURCE_TYPE_TEXTURE:
-            if (!runner->heap)
-                runner->heap = create_gpu_descriptor_heap(device,
-                        D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV, MAX_RESOURCE_DESCRIPTORS);
-
             if (params->desc.dimension == RESOURCE_DIMENSION_BUFFER)
             {
                 D3D12_SHADER_RESOURCE_VIEW_DESC srv_desc = { 0 };
@@ -229,10 +223,6 @@ static struct resource *d3d12_runner_create_resource(struct shader_runner *r, co
             break;
 
         case RESOURCE_TYPE_UAV:
-            if (!runner->heap)
-                runner->heap = create_gpu_descriptor_heap(device,
-                        D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV, MAX_RESOURCE_DESCRIPTORS);
-
             if (params->desc.dimension == RESOURCE_DIMENSION_BUFFER)
             {
                 D3D12_UNORDERED_ACCESS_VIEW_DESC uav_desc = { 0 };
@@ -294,6 +284,7 @@ static struct resource *d3d12_runner_create_resource(struct shader_runner *r, co
             break;
     }
 
+    free(resource_data);
     return &resource->r;
 }
 
@@ -305,16 +296,23 @@ static void d3d12_runner_destroy_resource(struct shader_runner *r, struct resour
     free(resource);
 }
 
+static const D3D12_DESCRIPTOR_RANGE_TYPE range_types[] =
+{
+    [VKD3D_SHADER_DESCRIPTOR_TYPE_SRV]     = D3D12_DESCRIPTOR_RANGE_TYPE_SRV,
+    [VKD3D_SHADER_DESCRIPTOR_TYPE_UAV]     = D3D12_DESCRIPTOR_RANGE_TYPE_UAV,
+    [VKD3D_SHADER_DESCRIPTOR_TYPE_CBV]     = D3D12_DESCRIPTOR_RANGE_TYPE_CBV,
+    [VKD3D_SHADER_DESCRIPTOR_TYPE_SAMPLER] = D3D12_DESCRIPTOR_RANGE_TYPE_SAMPLER,
+};
+
 static ID3D12RootSignature *d3d12_runner_create_root_signature(struct d3d12_shader_runner *runner,
         ID3D12CommandQueue *queue, ID3D12CommandAllocator *allocator,
         ID3D12GraphicsCommandList *command_list, unsigned int *uniform_index)
 {
     D3D12_ROOT_SIGNATURE_DESC root_signature_desc = {0};
+    D3D12_DESCRIPTOR_RANGE root_ranges[MAX_DESCRIPTORS];
     D3D12_ROOT_PARAMETER root_params[17], *root_param;
     D3D12_STATIC_SAMPLER_DESC static_samplers[7];
-    struct d3d12_resource *base_resource = NULL;
     ID3D12RootSignature *root_signature;
-    unsigned int slot;
     HRESULT hr;
     size_t i;
 
@@ -324,6 +322,32 @@ static ID3D12RootSignature *d3d12_runner_create_root_signature(struct d3d12_shad
     root_signature_desc.pStaticSamplers = static_samplers;
     root_signature_desc.Flags = D3D12_ROOT_SIGNATURE_FLAG_ALLOW_INPUT_ASSEMBLER_INPUT_LAYOUT;
 
+    /* Native d3d12 doesn't like an empty root descriptor table. */
+    if (runner->r.descriptor_count)
+    {
+        root_param = &root_params[root_signature_desc.NumParameters++];
+
+        root_param->ParameterType = D3D12_ROOT_PARAMETER_TYPE_DESCRIPTOR_TABLE;
+        root_param->ShaderVisibility = D3D12_SHADER_VISIBILITY_ALL;
+        root_param->DescriptorTable.NumDescriptorRanges = 0;
+        root_param->DescriptorTable.pDescriptorRanges = root_ranges;
+
+        for (i = 0; i < runner->r.descriptor_count; ++i)
+        {
+            D3D12_DESCRIPTOR_RANGE *range = &root_ranges[root_param->DescriptorTable.NumDescriptorRanges++];
+            struct descriptor_mapping *descriptor = &runner->r.descriptors[i];
+
+            range->RangeType = range_types[descriptor->type];
+            range->BaseShaderRegister = descriptor->register_idx;
+            range->RegisterSpace = descriptor->register_space;
+            range->NumDescriptors = descriptor->count;
+
+            range->OffsetInDescriptorsFromTableStart = descriptor->target_idx;
+            if (descriptor->type == VKD3D_SHADER_DESCRIPTOR_TYPE_UAV)
+                range->OffsetInDescriptorsFromTableStart += MAX_RESOURCES;
+        }
+    }
+
     if (runner->r.uniform_count)
     {
         *uniform_index = root_signature_desc.NumParameters++;
@@ -335,52 +359,6 @@ static ID3D12RootSignature *d3d12_runner_create_root_signature(struct d3d12_shad
         root_param->ShaderVisibility = D3D12_SHADER_VISIBILITY_ALL;
     }
 
-    for (i = 0; i < runner->r.resource_count; ++i)
-    {
-        struct d3d12_resource *resource = d3d12_resource(runner->r.resources[i]);
-        D3D12_DESCRIPTOR_RANGE *range;
-
-        switch (resource->r.desc.type)
-        {
-            case RESOURCE_TYPE_TEXTURE:
-            case RESOURCE_TYPE_UAV:
-                range = &resource->descriptor_range;
-
-                if (base_resource && resource->r.desc.type == base_resource->r.desc.type && resource->r.desc.slot == slot + 1)
-                {
-                    ++base_resource->descriptor_range.NumDescriptors;
-                    resource->descriptor_range.NumDescriptors = 0;
-                    ++slot;
-                    continue;
-                }
-
-                resource->root_index = root_signature_desc.NumParameters++;
-                root_param = &root_params[resource->root_index];
-                root_param->ParameterType = D3D12_ROOT_PARAMETER_TYPE_DESCRIPTOR_TABLE;
-                root_param->DescriptorTable.NumDescriptorRanges = 1;
-                root_param->DescriptorTable.pDescriptorRanges = range;
-                root_param->ShaderVisibility = D3D12_SHADER_VISIBILITY_ALL;
-
-                if (resource->r.desc.type == RESOURCE_TYPE_UAV)
-                    range->RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_UAV;
-                else
-                    range->RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_SRV;
-                range->NumDescriptors = 1;
-                range->BaseShaderRegister = resource->r.desc.slot;
-                range->RegisterSpace = 0;
-                range->OffsetInDescriptorsFromTableStart = 0;
-
-                base_resource = resource;
-                slot = resource->r.desc.slot;
-                break;
-
-            case RESOURCE_TYPE_RENDER_TARGET:
-            case RESOURCE_TYPE_DEPTH_STENCIL:
-            case RESOURCE_TYPE_VERTEX_BUFFER:
-                break;
-        }
-    }
-
     assert(root_signature_desc.NumParameters <= ARRAY_SIZE(root_params));
 
     for (i = 0; i < runner->r.sampler_count; ++i)
@@ -427,7 +405,6 @@ static bool d3d12_runner_dispatch(struct shader_runner *r, unsigned int x, unsig
     D3D12_SHADER_BYTECODE cs;
     ID3D10Blob *cs_code;
     HRESULT hr;
-    size_t i;
 
     cs_code = compile_hlsl(&runner->r, SHADER_TYPE_CS);
     todo_if(runner->r.is_todo && runner->r.minimum_shader_model < SHADER_MODEL_6_0) ok(cs_code, "Failed to compile shader.\n");
@@ -455,30 +432,10 @@ static bool d3d12_runner_dispatch(struct shader_runner *r, unsigned int x, unsig
     if (runner->r.uniform_count)
         ID3D12GraphicsCommandList_SetComputeRoot32BitConstants(command_list, uniform_index,
                 runner->r.uniform_count, runner->r.uniforms, 0);
-    for (i = 0; i < runner->r.resource_count; ++i)
-    {
-        struct d3d12_resource *resource = d3d12_resource(runner->r.resources[i]);
 
-        switch (resource->r.desc.type)
-        {
-            case RESOURCE_TYPE_TEXTURE:
-                if (resource->descriptor_range.NumDescriptors)
-                    ID3D12GraphicsCommandList_SetComputeRootDescriptorTable(command_list, resource->root_index,
-                            get_gpu_descriptor_handle(test_context, runner->heap, resource->r.desc.slot));
-                break;
-
-            case RESOURCE_TYPE_UAV:
-                if (resource->descriptor_range.NumDescriptors)
-                    ID3D12GraphicsCommandList_SetComputeRootDescriptorTable(command_list, resource->root_index,
-                            get_gpu_descriptor_handle(test_context, runner->heap, resource->r.desc.slot + MAX_RESOURCES));
-                break;
-
-            case RESOURCE_TYPE_RENDER_TARGET:
-            case RESOURCE_TYPE_DEPTH_STENCIL:
-            case RESOURCE_TYPE_VERTEX_BUFFER:
-                break;
-        }
-    }
+    if (r->descriptor_count)
+        ID3D12GraphicsCommandList_SetComputeRootDescriptorTable(command_list, 0,
+                get_gpu_descriptor_handle(test_context, runner->heap, 0));
 
     ID3D12GraphicsCommandList_SetPipelineState(command_list, pso);
     ID3D12GraphicsCommandList_Dispatch(command_list, x, y, z);
@@ -807,8 +764,8 @@ static ID3D12PipelineState *create_pipeline_device2(struct d3d12_shader_runner *
     return pso;
 }
 
-static bool d3d12_runner_draw(struct shader_runner *r,
-        D3D_PRIMITIVE_TOPOLOGY primitive_topology, unsigned int vertex_count, unsigned int instance_count)
+static bool d3d12_runner_draw(struct shader_runner *r, D3D_PRIMITIVE_TOPOLOGY primitive_topology,
+        unsigned int vertex_count, unsigned int instance_count)
 {
     struct d3d12_shader_runner *runner = d3d12_shader_runner(r);
     struct test_context *test_context = &runner->test_context;
@@ -895,8 +852,7 @@ static bool d3d12_runner_draw(struct shader_runner *r,
     if (runner->r.uniform_count)
         ID3D12GraphicsCommandList_SetGraphicsRoot32BitConstants(command_list, uniform_index,
                 runner->r.uniform_count, runner->r.uniforms, 0);
-    if (runner->heap)
-        ID3D12GraphicsCommandList_SetDescriptorHeaps(command_list, 1, &runner->heap);
+    ID3D12GraphicsCommandList_SetDescriptorHeaps(command_list, 1, &runner->heap);
     for (i = 0; i < runner->r.resource_count; ++i)
     {
         struct d3d12_resource *resource = d3d12_resource(runner->r.resources[i]);
@@ -921,18 +877,6 @@ static bool d3d12_runner_draw(struct shader_runner *r,
                     fb_height = resource->r.desc.height;
                 break;
 
-            case RESOURCE_TYPE_TEXTURE:
-                if (resource->descriptor_range.NumDescriptors)
-                    ID3D12GraphicsCommandList_SetGraphicsRootDescriptorTable(command_list, resource->root_index,
-                            get_gpu_descriptor_handle(test_context, runner->heap, resource->r.desc.slot));
-                break;
-
-            case RESOURCE_TYPE_UAV:
-                if (resource->descriptor_range.NumDescriptors)
-                    ID3D12GraphicsCommandList_SetGraphicsRootDescriptorTable(command_list, resource->root_index,
-                            get_gpu_descriptor_handle(test_context, runner->heap, resource->r.desc.slot + MAX_RESOURCES));
-                break;
-
             case RESOURCE_TYPE_VERTEX_BUFFER:
                 vbv.BufferLocation = ID3D12Resource_GetGPUVirtualAddress(resource->resource);
                 vbv.StrideInBytes = get_vb_stride(&runner->r, resource->r.desc.slot);
@@ -940,9 +884,17 @@ static bool d3d12_runner_draw(struct shader_runner *r,
 
                 ID3D12GraphicsCommandList_IASetVertexBuffers(command_list, resource->r.desc.slot, 1, &vbv);
                 break;
+
+            case RESOURCE_TYPE_TEXTURE:
+            case RESOURCE_TYPE_UAV:
+                break;
         }
     }
 
+    if (r->descriptor_count)
+        ID3D12GraphicsCommandList_SetGraphicsRootDescriptorTable(command_list, 0,
+                get_gpu_descriptor_handle(test_context, runner->heap, 0));
+
     set_rect(&test_context->scissor_rect, 0, 0, fb_width, fb_height);
     set_viewport(&test_context->viewport, 0.0f, 0.0f, fb_width, fb_height, 0.0f, 1.0f);
     viewports[0] = test_context->viewport;
@@ -1110,7 +1062,10 @@ static void d3d12_runner_init_caps(struct d3d12_shader_runner *runner,
     runner->caps.compiler = using_dxcompiler ? "dxcompiler" : HLSL_COMPILER;
     runner->caps.minimum_shader_model = minimum_shader_model;
     runner->caps.maximum_shader_model = maximum_shader_model;
+    if (is_cull_distance_supported(device))
+        runner->caps.shader_caps[SHADER_CAP_CULL_DISTANCE] = true;
     runner->caps.shader_caps[SHADER_CAP_DEPTH_BOUNDS] = options2.DepthBoundsTestSupported;
+    runner->caps.shader_caps[SHADER_CAP_DESCRIPTORS] = true;
     runner->caps.shader_caps[SHADER_CAP_FLOAT64] = options.DoublePrecisionFloatShaderOps;
     if (is_geometry_shader_supported(device))
         runner->caps.shader_caps[SHADER_CAP_GEOMETRY_SHADER] = true;
@@ -1122,28 +1077,43 @@ static void d3d12_runner_init_caps(struct d3d12_shader_runner *runner,
     runner->caps.shader_caps[SHADER_CAP_WAVE_OPS] = options1.WaveOps;
     runner->caps.shader_caps[SHADER_CAP_NATIVE_16_BIT] = options4.Native16BitShaderOpsSupported;
 
-    runner->caps.tag_count = 0;
-    runner->caps.tags[runner->caps.tag_count++] = "d3d12";
-    if (is_mvk_device(device))
-    {
-        runner->caps.tags[runner->caps.tag_count++] = "mvk";
-        if (is_mvk_device_lt(device, 1, 2, 11))
-            runner->caps.tags[runner->caps.tag_count++] = "mvk<1.2.11";
-    }
-    else
-    {
-        if (is_llvmpipe_device(device))
-            runner->caps.tags[runner->caps.tag_count++] = "llvmpipe";
-        if (is_mesa_device_lt(device, 23, 3, 0))
-            runner->caps.tags[runner->caps.tag_count++] = "mesa<23.3";
-        if (test_options.use_warp_device)
-            runner->caps.tags[runner->caps.tag_count++] = "warp";
-    }
+    runner->caps.tags[SHADER_RUNNER_TAG_D3D12].present = true;
+
+    runner->caps.tags[SHADER_RUNNER_TAG_MVK].present = get_mvk_device_version(device,
+            &runner->caps.tags[SHADER_RUNNER_TAG_MVK].major, &runner->caps.tags[SHADER_RUNNER_TAG_MVK].minor,
+            &runner->caps.tags[SHADER_RUNNER_TAG_MVK].patch);
+
+    if (is_llvmpipe_device(device))
+        runner->caps.tags[SHADER_RUNNER_TAG_LLVMPIPE].present = true;
+
+    if (is_intel_device(device))
+        runner->caps.tags[SHADER_RUNNER_TAG_INTEL].present = true;
+
+    runner->caps.tags[SHADER_RUNNER_TAG_MESA].present = get_mesa_device_version(device,
+            &runner->caps.tags[SHADER_RUNNER_TAG_MESA].major, &runner->caps.tags[SHADER_RUNNER_TAG_MESA].minor,
+            &runner->caps.tags[SHADER_RUNNER_TAG_MESA].patch);
+
+    runner->caps.tags[SHADER_RUNNER_TAG_LLVM].present = get_llvm_device_version(device,
+            &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].major, &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].minor,
+            &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].patch);
+
+    if (test_options.use_warp_device)
+        runner->caps.tags[SHADER_RUNNER_TAG_WARP].present = true;
+
+    if (is_nvidia_windows_device(device) || is_nvidia_device(device))
+        runner->caps.tags[SHADER_RUNNER_TAG_NVIDIA].present = true;
+
+#ifdef _WIN32
+    runner->caps.tags[SHADER_RUNNER_TAG_WINDOWS].present = true;
+#endif
 
     for (unsigned int i = 0; i < ARRAY_SIZE(formats); ++i)
     {
         runner->caps.format_caps[formats[i]] = get_format_support(device, formats[i]);
     }
+
+    runner->heap = create_gpu_descriptor_heap(device,
+            D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV, MAX_RESOURCE_DESCRIPTORS);
 }
 
 static bool device_supports_shader_model_6_0(ID3D12Device *device)
