@@ -75,38 +75,20 @@ static HRESULT set_viewport(IDirect3DDevice9 *device, unsigned int x, unsigned i
     return IDirect3DDevice9_SetViewport(device, &vp);
 }
 
-static void init_adapter_info(void)
-{
-    D3DADAPTER_IDENTIFIER9 identifier;
-    IDirect3D9 *d3d;
-    HRESULT hr;
-
-    d3d = pDirect3DCreate9(D3D_SDK_VERSION);
-    ok(!!d3d, "Failed to create a D3D object.\n");
-
-    hr = IDirect3D9_GetAdapterIdentifier(d3d, test_options.adapter_idx, 0, &identifier);
-    ok(hr == S_OK, "Failed to get adapter identifier, hr %#lx.\n", hr);
-
-    trace("Driver string: %s.\n", identifier.Driver);
-    trace("Device: %s, %04lx:%04lx.\n", identifier.Description, identifier.VendorId, identifier.DeviceId);
-
-    if (identifier.VendorId == 0x1414 && identifier.DeviceId == 0x008c)
-        trace("Using WARP device.\n");
-
-    IDirect3D9_Release(d3d);
-}
+static const BOOL windowed = TRUE;
 
 static bool init_test_context(struct d3d9_shader_runner *runner)
 {
     D3DPRESENT_PARAMETERS present_parameters =
     {
-        .Windowed = TRUE,
+        .Windowed = windowed,
         .SwapEffect = D3DSWAPEFFECT_DISCARD,
         .BackBufferWidth = RENDER_TARGET_WIDTH,
         .BackBufferHeight = RENDER_TARGET_HEIGHT,
         .BackBufferFormat = D3DFMT_A8R8G8B8,
     };
     RECT rect = {0, 0, RENDER_TARGET_WIDTH, RENDER_TARGET_HEIGHT};
+    D3DADAPTER_IDENTIFIER9 identifier;
     IDirect3D9 *d3d;
     D3DCAPS9 caps;
     HRESULT hr;
@@ -121,6 +103,12 @@ static bool init_test_context(struct d3d9_shader_runner *runner)
     d3d = pDirect3DCreate9(D3D_SDK_VERSION);
     ok(!!d3d, "Failed to create a D3D object.\n");
 
+    hr = IDirect3D9_GetAdapterIdentifier(d3d, test_options.adapter_idx, 0, &identifier);
+    ok(hr == S_OK, "Failed to get adapter identifier, hr %#lx.\n", hr);
+
+    trace("Driver string: %s.\n", identifier.Driver);
+    trace("Device: %s, %04lx:%04lx.\n", identifier.Description, identifier.VendorId, identifier.DeviceId);
+
     present_parameters.hDeviceWindow = runner->window;
 
     hr = IDirect3D9_CreateDevice(d3d, test_options.adapter_idx, D3DDEVTYPE_HAL, runner->window,
@@ -151,6 +139,18 @@ static bool init_test_context(struct d3d9_shader_runner *runner)
     runner->caps.shader_caps[SHADER_CAP_FOG] = true;
     runner->caps.shader_caps[SHADER_CAP_POINT_SIZE] = true;
 
+    runner->caps.tags[SHADER_RUNNER_TAG_WINDOWS].present = true;
+
+    if (identifier.VendorId == 0x1414 && identifier.DeviceId == 0x008c)
+    {
+        trace("Using WARP device.\n");
+        runner->caps.tags[SHADER_RUNNER_TAG_WARP].present = true;
+    }
+    else if (identifier.VendorId == 0x10de)
+    {
+        runner->caps.tags[SHADER_RUNNER_TAG_NVIDIA].present = true;
+    }
+
     return true;
 }
 
@@ -191,6 +191,7 @@ static struct resource *d3d9_runner_create_resource(struct shader_runner *r, con
     struct d3d9_shader_runner *runner = d3d9_shader_runner(r);
     IDirect3DDevice9 *device = runner->device;
     struct d3d9_resource *resource;
+    D3DMULTISAMPLE_TYPE ms_type;
     D3DFORMAT format;
     HRESULT hr;
     void *data;
@@ -216,8 +217,31 @@ static struct resource *d3d9_runner_create_resource(struct shader_runner *r, con
     switch (params->desc.type)
     {
         case RESOURCE_TYPE_RENDER_TARGET:
+            ms_type = D3DMULTISAMPLE_NONE;
+            if (params->desc.sample_count > 1)
+            {
+                IDirect3D9 *d3d;
+
+                assert(params->desc.sample_count <= 16);
+                ms_type = (D3DMULTISAMPLE_TYPE)params->desc.sample_count;
+
+                d3d = pDirect3DCreate9(D3D_SDK_VERSION);
+                ok(!!d3d, "Failed to create a D3D object.\n");
+
+                hr = IDirect3D9_CheckDeviceMultiSampleType(d3d, test_options.adapter_idx, D3DDEVTYPE_HAL,
+                        format, windowed, ms_type, NULL);
+                IDirect3D9_Release(d3d);
+
+                if (FAILED(hr))
+                {
+                    fatal_error("Format #%x with sample count %u is not supported.\n",
+                            format, params->desc.sample_count);
+                    break;
+                }
+            }
+
             hr = IDirect3DDevice9_CreateRenderTarget(device, params->desc.width, params->desc.height,
-                    format, D3DMULTISAMPLE_NONE, 0, FALSE, &resource->surface, NULL);
+                    format, ms_type, 0, FALSE, &resource->surface, NULL);
             ok(hr == D3D_OK, "Failed to create render target, hr %#lx.\n", hr);
             break;
 
@@ -240,6 +264,9 @@ static struct resource *d3d9_runner_create_resource(struct shader_runner *r, con
                         params->desc.level_count, 0, format, D3DPOOL_MANAGED, &resource->texture, NULL);
                 ok(hr == D3D_OK, "Failed to create texture, hr %#lx.\n", hr);
 
+                if (!params->data_size)
+                    break;
+
                 for (unsigned int level = 0; level < params->desc.level_count; ++level)
                 {
                     unsigned int level_width = get_level_dimension(params->desc.width, level);
@@ -540,13 +567,13 @@ static bool d3d9_runner_draw(struct shader_runner *r,
         ok(hr == D3D_OK, "Failed to set sampler state, hr %#lx.\n", hr);
 
         hr = IDirect3DDevice9_SetSamplerState(device, sampler->slot, D3DSAMP_MINFILTER,
-                (sampler->filter & 0x1) ? D3DTEXF_LINEAR : D3DTEXF_POINT);
+                (sampler->filter & 0x10) ? D3DTEXF_LINEAR : D3DTEXF_POINT);
         ok(hr == D3D_OK, "Failed to set sampler state, hr %#lx.\n", hr);
         hr = IDirect3DDevice9_SetSamplerState(device, sampler->slot, D3DSAMP_MAGFILTER,
                 (sampler->filter & 0x4) ? D3DTEXF_LINEAR : D3DTEXF_POINT);
         ok(hr == D3D_OK, "Failed to set sampler state, hr %#lx.\n", hr);
         hr = IDirect3DDevice9_SetSamplerState(device, sampler->slot, D3DSAMP_MIPFILTER,
-                (sampler->filter & 0x10) ? D3DTEXF_LINEAR : D3DTEXF_POINT);
+                (sampler->filter & 0x1) ? D3DTEXF_LINEAR : D3DTEXF_POINT);
         ok(hr == D3D_OK, "Failed to set sampler state, hr %#lx.\n", hr);
     }
 
@@ -604,6 +631,28 @@ static bool d3d9_runner_draw(struct shader_runner *r,
     hr = IDirect3DDevice9_SetRenderState(device, D3DRS_FOGDENSITY, float_to_int(runner->r.fog_density));
     ok(hr == D3D_OK, "Failed to set render state, hr %#lx.\n", hr);
 
+    for (unsigned int i = 0; i < 8; ++i)
+    {
+        hr = IDirect3DDevice9_SetTextureStageState(device, i, D3DTSS_BUMPENVMAT00,
+                float_to_int(runner->r.bump[i].matrix[0][0]));
+        ok(hr == D3D_OK, "Failed to set texture state, hr %#lx.\n", hr);
+        hr = IDirect3DDevice9_SetTextureStageState(device, i, D3DTSS_BUMPENVMAT01,
+                float_to_int(runner->r.bump[i].matrix[0][1]));
+        ok(hr == D3D_OK, "Failed to set texture state, hr %#lx.\n", hr);
+        hr = IDirect3DDevice9_SetTextureStageState(device, i, D3DTSS_BUMPENVMAT10,
+                float_to_int(runner->r.bump[i].matrix[1][0]));
+        ok(hr == D3D_OK, "Failed to set texture state, hr %#lx.\n", hr);
+        hr = IDirect3DDevice9_SetTextureStageState(device, i, D3DTSS_BUMPENVMAT11,
+                float_to_int(runner->r.bump[i].matrix[1][1]));
+        ok(hr == D3D_OK, "Failed to set texture state, hr %#lx.\n", hr);
+        hr = IDirect3DDevice9_SetTextureStageState(device, i, D3DTSS_BUMPENVLSCALE,
+                float_to_int(runner->r.bump[i].luminance_scale));
+        ok(hr == D3D_OK, "Failed to set texture state, hr %#lx.\n", hr);
+        hr = IDirect3DDevice9_SetTextureStageState(device, i, D3DTSS_BUMPENVLOFFSET,
+                float_to_int(runner->r.bump[i].luminance_offset));
+        ok(hr == D3D_OK, "Failed to set texture state, hr %#lx.\n", hr);
+    }
+
     hr = IDirect3DDevice9_CreateVertexDeclaration(device, decl_elements, &vertex_declaration);
     ok(hr == D3D_OK, "Failed to create vertex declaration, hr %#lx.\n", hr);
     hr = IDirect3DDevice9_CreateVertexShader(device, ID3D10Blob_GetBufferPointer(vs_code), &vs);
@@ -652,11 +701,6 @@ static bool d3d9_runner_draw(struct shader_runner *r,
     return true;
 }
 
-static bool d3d9_runner_copy(struct shader_runner *r, struct resource *src, struct resource *dst)
-{
-    return false;
-}
-
 struct d3d9_resource_readback
 {
     struct resource_readback rb;
@@ -669,6 +713,8 @@ static struct resource_readback *d3d9_runner_get_resource_readback(struct shader
     struct d3d9_shader_runner *runner = d3d9_shader_runner(r);
     struct d3d9_resource_readback *rb = malloc(sizeof(*rb));
     struct d3d9_resource *resource = d3d9_resource(res);
+    IDirect3DSurface9 *surface = resource->surface;
+    IDirect3DSurface9 *resolved_surface = NULL;
     D3DLOCKED_RECT map_desc;
     D3DSURFACE_DESC desc;
     HRESULT hr;
@@ -677,15 +723,30 @@ static struct resource_readback *d3d9_runner_get_resource_readback(struct shader
     if (sub_resource_idx)
         fatal_error("Unsupported sub-resource index %u.\n", sub_resource_idx);
 
-    hr = IDirect3DSurface9_GetDesc(resource->surface, &desc);
+    hr = IDirect3DSurface9_GetDesc(surface, &desc);
     ok(hr == D3D_OK, "Failed to get surface desc, hr %#lx.\n", hr);
     hr = IDirect3DDevice9Ex_CreateOffscreenPlainSurface(runner->device, desc.Width,
             desc.Height, desc.Format, D3DPOOL_SYSTEMMEM, &rb->surface, NULL);
     ok(hr == D3D_OK, "Failed to create surface, hr %#lx.\n", hr);
 
-    hr = IDirect3DDevice9Ex_GetRenderTargetData(runner->device, resource->surface, rb->surface);
+    if (desc.MultiSampleType != D3DMULTISAMPLE_NONE)
+    {
+        hr = IDirect3DDevice9_CreateRenderTarget(runner->device, desc.Width, desc.Height, desc.Format,
+                D3DMULTISAMPLE_NONE, 0, FALSE, &resolved_surface, NULL);
+        ok(hr == D3D_OK, "Failed to create resolved render target, hr %#lx.\n", hr);
+
+        hr = IDirect3DDevice9_StretchRect(runner->device, surface, NULL, resolved_surface, NULL, D3DTEXF_NONE);
+        ok(hr == D3D_OK, "Failed to resolve multisampled surface, hr %#lx.\n", hr);
+
+        surface = resolved_surface;
+    }
+
+    hr = IDirect3DDevice9Ex_GetRenderTargetData(runner->device, surface, rb->surface);
     ok(hr == D3D_OK, "Failed to get render target data, hr %#lx.\n", hr);
 
+    if (resolved_surface)
+        IDirect3DSurface9_Release(resolved_surface);
+
     hr = IDirect3DSurface9_LockRect(rb->surface, &map_desc, NULL, D3DLOCK_READONLY);
     ok(hr == D3D_OK, "Failed to lock surface, hr %#lx.\n", hr);
 
@@ -706,6 +767,37 @@ static void d3d9_runner_release_readback(struct shader_runner *r, struct resourc
     free(d3d9_rb);
 }
 
+static bool d3d9_runner_copy(struct shader_runner *r, struct resource *s, struct resource *d)
+{
+    struct d3d9_resource *src = d3d9_resource(s);
+    struct d3d9_resource *dst = d3d9_resource(d);
+
+    if (src->r.desc.type == RESOURCE_TYPE_RENDER_TARGET && dst->r.desc.type == RESOURCE_TYPE_TEXTURE)
+    {
+        /* We create our textures with MANAGED, and unfortunately there's no
+         * way to directly blit into a managed texture. */
+        struct resource_readback *rb;
+        D3DLOCKED_RECT map_desc;
+        HRESULT hr;
+
+        rb = d3d9_runner_get_resource_readback(r, s, 0);
+
+        hr = IDirect3DTexture9_LockRect(dst->texture, 0, &map_desc, NULL, 0);
+        ok(hr == S_OK, "Failed to map texture, hr %#lx.\n", hr);
+        for (unsigned int y = 0; y < src->r.desc.height; ++y)
+            memcpy(&((char *)map_desc.pBits)[y * map_desc.Pitch], &((char *)rb->data)[y * rb->row_pitch],
+                    src->r.desc.width * src->r.desc.texel_size);
+        hr = IDirect3DTexture9_UnlockRect(dst->texture, 0);
+        ok(hr == S_OK, "Failed to unmap texture, hr %#lx.\n", hr);
+
+        d3d9_runner_release_readback(r, rb);
+
+        return true;
+    }
+
+    return false;
+}
+
 static const struct shader_runner_ops d3d9_runner_ops =
 {
     .create_resource = d3d9_runner_create_resource,
@@ -732,7 +824,6 @@ void run_shader_tests_d3d9(void)
 
     pDirect3DCreate9 = (void *)GetProcAddress(d3d9_module, "Direct3DCreate9");
 
-    init_adapter_info();
     if (init_test_context(&runner))
     {
         run_shader_tests(&runner.r, &runner.caps, &d3d9_runner_ops, NULL);
