@@ -83,6 +83,8 @@ struct gl_runner
     struct vkd3d_shader_combined_resource_sampler *combined_samplers;
     unsigned int combined_sampler_count;
     enum shading_language language;
+
+    uint32_t clip_distance_mask[SHADER_TYPE_COUNT];
 };
 
 static struct gl_runner *gl_runner(struct shader_runner *r)
@@ -153,6 +155,8 @@ static bool check_gl_extensions(struct gl_runner *runner)
 
     if (check_gl_extension("GL_EXT_depth_bounds_test", count))
         runner->caps.shader_caps[SHADER_CAP_DEPTH_BOUNDS] = true;
+    if (check_gl_extension("GL_ARB_cull_distance", count))
+        runner->caps.shader_caps[SHADER_CAP_CULL_DISTANCE] = true;
     if (check_gl_extension("GL_ARB_gpu_shader_fp64", count))
         runner->caps.shader_caps[SHADER_CAP_FLOAT64] = true;
     if (check_gl_extension("GL_ARB_gpu_shader_int64", count))
@@ -247,6 +251,16 @@ static uint32_t get_format_support(struct gl_runner *runner, enum DXGI_FORMAT fo
     return ret;
 }
 
+static bool get_mesa_version(uint32_t *major, uint32_t *minor, uint32_t *patch)
+{
+    return find_version_string((const char *)glGetString(GL_VERSION), "Mesa ", major, minor, patch);
+}
+
+static bool get_llvm_version(uint32_t *major, uint32_t *minor, uint32_t *patch)
+{
+    return find_version_string((const char *)glGetString(GL_RENDERER), "LLVM ", major, minor, patch);
+}
+
 static bool gl_runner_init(struct gl_runner *runner, enum shading_language language)
 {
     PFNEGLQUERYDEVICESEXTPROC eglQueryDevicesEXT;
@@ -368,12 +382,21 @@ static bool gl_runner_init(struct gl_runner *runner, enum shading_language langu
         runner->caps.maximum_shader_model = SHADER_MODEL_5_1;
         runner->caps.shader_caps[SHADER_CAP_GEOMETRY_SHADER] = true;
 
-        runner->caps.tag_count = 0;
-        runner->caps.tags[runner->caps.tag_count++] = "opengl";
+        runner->caps.tags[SHADER_RUNNER_TAG_OPENGL].present = true;
+
         if (runner->language == GLSL)
-            runner->caps.tags[runner->caps.tag_count++] = "glsl";
+            runner->caps.tags[SHADER_RUNNER_TAG_GLSL].present = true;
+
         if (strncmp((const char *)glGetString(GL_RENDERER), "llvmpipe ", 9) == 0)
-            runner->caps.tags[runner->caps.tag_count++] = "llvmpipe";
+            runner->caps.tags[SHADER_RUNNER_TAG_LLVMPIPE].present = true;
+
+        runner->caps.tags[SHADER_RUNNER_TAG_MESA].present = get_mesa_version(
+            &runner->caps.tags[SHADER_RUNNER_TAG_MESA].major, &runner->caps.tags[SHADER_RUNNER_TAG_MESA].minor,
+            &runner->caps.tags[SHADER_RUNNER_TAG_MESA].patch);
+
+        runner->caps.tags[SHADER_RUNNER_TAG_LLVM].present = get_llvm_version(
+            &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].major, &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].minor,
+            &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].patch);
 
         glGetIntegerv(GL_NUM_EXTENSIONS, &extension_count);
         if (check_gl_extension("GL_ARB_internalformat_query2", extension_count))
@@ -625,6 +648,25 @@ static void gl_runner_destroy_resource(struct shader_runner *r, struct resource
     free(resource);
 }
 
+static uint32_t get_clip_distance_mask(const struct vkd3d_shader_signature *output)
+{
+    const struct vkd3d_shader_signature_element *e;
+    uint32_t mask = 0;
+    unsigned int i;
+
+    for (i = 0; i < output->element_count; ++i)
+    {
+        e = &output->elements[i];
+        if (e->sysval_semantic != VKD3D_SHADER_SV_CLIP_DISTANCE)
+            continue;
+
+        assert(e->semantic_index < 32u);
+        mask |= 1u << e->semantic_index;
+    }
+
+    return mask;
+}
+
 static bool compile_shader(struct gl_runner *runner, enum shader_type shader_type,
         ID3DBlob *blob, struct vkd3d_shader_code *out)
 {
@@ -636,6 +678,7 @@ static bool compile_shader(struct gl_runner *runner, enum shader_type shader_typ
     struct vkd3d_shader_scan_hull_shader_tessellation_info tessellation_info;
     struct vkd3d_shader_spirv_domain_shader_target_info domain_info;
     struct vkd3d_shader_combined_resource_sampler *sampler;
+    struct vkd3d_shader_scan_signature_info signature_info;
     enum vkd3d_shader_spirv_extension spirv_extensions[1];
     struct vkd3d_shader_resource_binding *binding;
     struct vkd3d_shader_parameter parameters[1];
@@ -645,7 +688,7 @@ static bool compile_shader(struct gl_runner *runner, enum shader_type shader_typ
 
     const struct vkd3d_shader_compile_option options[] =
     {
-        {VKD3D_SHADER_COMPILE_OPTION_API_VERSION, VKD3D_SHADER_API_VERSION_1_17},
+        {VKD3D_SHADER_COMPILE_OPTION_API_VERSION, VKD3D_SHADER_API_VERSION_CURRENT},
         {VKD3D_SHADER_COMPILE_OPTION_FRAGMENT_COORDINATE_ORIGIN,
                 VKD3D_SHADER_COMPILE_OPTION_FRAGMENT_COORDINATE_ORIGIN_LOWER_LEFT},
         {VKD3D_SHADER_COMPILE_OPTION_FEATURE, shader_runner_caps_get_feature_flags(&runner->caps)},
@@ -661,7 +704,10 @@ static bool compile_shader(struct gl_runner *runner, enum shader_type shader_typ
     info.log_level = VKD3D_SHADER_LOG_WARNING;
 
     combined_sampler_info.type = VKD3D_SHADER_STRUCTURE_TYPE_SCAN_COMBINED_RESOURCE_SAMPLER_INFO;
-    combined_sampler_info.next = &tessellation_info;
+    combined_sampler_info.next = &signature_info;
+
+    signature_info.type = VKD3D_SHADER_STRUCTURE_TYPE_SCAN_SIGNATURE_INFO;
+    signature_info.next = &tessellation_info;
 
     tessellation_info.type = VKD3D_SHADER_STRUCTURE_TYPE_SCAN_HULL_SHADER_TESSELLATION_INFO;
     tessellation_info.next = NULL;
@@ -703,6 +749,9 @@ static bool compile_shader(struct gl_runner *runner, enum shader_type shader_typ
     }
     vkd3d_shader_free_scan_combined_resource_sampler_info(&combined_sampler_info);
 
+    runner->clip_distance_mask[shader_type] = get_clip_distance_mask(&signature_info.output);
+    vkd3d_shader_free_scan_signature_info(&signature_info);
+
     if (runner->language == SPIR_V)
     {
         info.next = &spirv_info;
@@ -1026,10 +1075,12 @@ static GLenum get_texture_filter_mag_gl(D3D12_FILTER filter)
 
 static GLenum get_texture_filter_min_gl(D3D12_FILTER filter)
 {
-    if (filter & 0x1)
-        return filter & 0x10 ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR_MIPMAP_NEAREST;
+    bool min_linear = filter & 0x10, mip_linear = filter & 0x1;
+
+    if (min_linear)
+        return mip_linear ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR_MIPMAP_NEAREST;
     else
-        return filter & 0x10 ? GL_NEAREST_MIPMAP_LINEAR : GL_NEAREST_MIPMAP_NEAREST;
+        return mip_linear ? GL_NEAREST_MIPMAP_LINEAR : GL_NEAREST_MIPMAP_NEAREST;
 }
 
 static GLenum get_compare_op_gl(D3D12_COMPARISON_FUNC op)
@@ -1204,15 +1255,16 @@ static bool gl_runner_draw(struct shader_runner *r,
     struct vkd3d_shader_signature vs_input_signature;
     struct gl_runner *runner = gl_runner(r);
     struct vkd3d_shader_code vs_dxbc;
+    uint32_t clip_distance_mask, map;
     uint8_t *attribute_offsets[32];
     struct
     {
         GLuint id;
         GLsizei stride;
     } vbo_info[MAX_RESOURCES];
+    GLint max_clip_distance_count;
     GLuint program_id, ubo_id = 0;
     ID3D10Blob *vs_blob;
-    uint32_t map;
     int ret;
     struct
     {
@@ -1399,6 +1451,35 @@ static bool gl_runner_draw(struct shader_runner *r,
         runner->attribute_map |= attribute_idx;
     }
     vkd3d_shader_free_shader_signature(&vs_input_signature);
+
+    for (i = 0, clip_distance_mask = 0; i < SHADER_TYPE_COUNT; ++i)
+    {
+        clip_distance_mask |= runner->clip_distance_mask[i];
+    }
+
+    glGetIntegerv(GL_MAX_CLIP_DISTANCES, &max_clip_distance_count);
+    if (clip_distance_mask & ~((1u << max_clip_distance_count) - 1))
+        trace("Warning: Not enough clip distances supported (maximum is %d).\n", max_clip_distance_count);
+
+    /* OpenGL expects all declared gl_ClipDistance[] arrays to be the same
+     * size across shaders. */
+    for (i = 0; i < SHADER_TYPE_COUNT; ++i)
+    {
+        if (!runner->clip_distance_mask[i])
+            continue;
+        if (runner->clip_distance_mask[i] != clip_distance_mask)
+            trace("Warning: Shader type %#x has clip distance mask %#x (subset of the combined %#x).\n",
+                    i, runner->clip_distance_mask[i], clip_distance_mask);
+    }
+
+    for (i = 0; i < max_clip_distance_count; ++i)
+    {
+        if (clip_distance_mask & (1u << i))
+            glEnable(GL_CLIP_DISTANCE0 + i);
+        else
+            glDisable(GL_CLIP_DISTANCE0 + i);
+    }
+
     map &= ~runner->attribute_map;
     for (attribute_idx = 0; map; ++attribute_idx, map >>= 1)
     {
