@@ -58,6 +58,7 @@ struct metal_runner
 
     ID3D10Blob *d3d_blobs[SHADER_TYPE_COUNT];
     struct vkd3d_shader_scan_signature_info signatures[SHADER_TYPE_COUNT];
+    struct vkd3d_shader_scan_thread_group_size_info thread_group_size;
 };
 
 static MTLPixelFormat get_metal_pixel_format(DXGI_FORMAT format)
@@ -385,15 +386,15 @@ static bool compile_shader(struct metal_runner *runner, enum shader_type type, s
 {
     struct vkd3d_shader_interface_info interface_info = {.type = VKD3D_SHADER_STRUCTURE_TYPE_INTERFACE_INFO};
     struct vkd3d_shader_compile_info info = {.type = VKD3D_SHADER_STRUCTURE_TYPE_COMPILE_INFO};
-    struct vkd3d_shader_resource_binding bindings[MAX_RESOURCES + MAX_SAMPLERS + 1 /* CBV */];
+    struct vkd3d_shader_resource_binding bindings[MAX_DESCRIPTORS + MAX_SAMPLERS + 1 /* CBV */];
     struct vkd3d_shader_resource_binding *binding;
-    unsigned int i;
+    unsigned int i, binding_idx = 0;
     char *messages;
     int ret;
 
     const struct vkd3d_shader_compile_option options[] =
     {
-        {VKD3D_SHADER_COMPILE_OPTION_API_VERSION, VKD3D_SHADER_API_VERSION_1_17},
+        {VKD3D_SHADER_COMPILE_OPTION_API_VERSION, VKD3D_SHADER_API_VERSION_CURRENT},
         {VKD3D_SHADER_COMPILE_OPTION_FEATURE, shader_runner_caps_get_feature_flags(&runner->caps)},
     };
 
@@ -421,55 +422,29 @@ static bool compile_shader(struct metal_runner *runner, enum shader_type type, s
         binding->shader_visibility = VKD3D_SHADER_VISIBILITY_ALL;
         binding->flags = VKD3D_SHADER_BINDING_FLAG_BUFFER;
         binding->binding.set = 0;
-        binding->binding.binding = interface_info.binding_count;
+        binding->binding.binding = binding_idx;
         binding->binding.count = 1;
+
         ++interface_info.binding_count;
+        ++binding_idx;
     }
 
-    for (i = 0; i < runner->r.resource_count; ++i)
+    for (i = 0; i < runner->r.descriptor_count; ++i)
     {
-        const struct metal_resource *resource = metal_resource(runner->r.resources[i]);
+        const struct descriptor_mapping *descriptor = &runner->r.descriptors[i];
 
-        switch (resource->r.desc.type)
-        {
-            case RESOURCE_TYPE_TEXTURE:
-                binding = &bindings[interface_info.binding_count];
-                binding->type = VKD3D_SHADER_DESCRIPTOR_TYPE_SRV;
-                binding->register_space = 0;
-                binding->register_index = resource->r.desc.slot;
-                binding->shader_visibility = VKD3D_SHADER_VISIBILITY_ALL;
-                if (resource->r.desc.dimension == RESOURCE_DIMENSION_BUFFER)
-                    binding->flags = VKD3D_SHADER_BINDING_FLAG_BUFFER;
-                else
-                    binding->flags = VKD3D_SHADER_BINDING_FLAG_IMAGE;
-                binding->binding.set = 0;
-                binding->binding.binding = interface_info.binding_count;
-                binding->binding.count = 1;
-                ++interface_info.binding_count;
-                break;
+        binding = &bindings[interface_info.binding_count];
+        binding->type = descriptor->type;
+        binding->register_index = descriptor->register_idx;
+        binding->register_space = descriptor->register_space;
+        binding->shader_visibility = VKD3D_SHADER_VISIBILITY_ALL;
+        binding->flags = VKD3D_SHADER_BINDING_FLAG_BUFFER | VKD3D_SHADER_BINDING_FLAG_IMAGE;
+        binding->binding.set = 0;
+        binding->binding.binding = binding_idx;
+        binding->binding.count = descriptor->count;
 
-            case RESOURCE_TYPE_UAV:
-                binding = &bindings[interface_info.binding_count];
-                binding->type = VKD3D_SHADER_DESCRIPTOR_TYPE_UAV;
-                binding->register_space = 0;
-                binding->register_index = resource->r.desc.slot;
-                binding->shader_visibility = VKD3D_SHADER_VISIBILITY_ALL;
-                if (resource->r.desc.dimension == RESOURCE_DIMENSION_BUFFER)
-                    binding->flags = VKD3D_SHADER_BINDING_FLAG_BUFFER;
-                else
-                    binding->flags = VKD3D_SHADER_BINDING_FLAG_IMAGE;
-                binding->binding.set = 0;
-                binding->binding.binding = interface_info.binding_count;
-                binding->binding.count = 1;
-                ++interface_info.binding_count;
-                break;
-
-            case RESOURCE_TYPE_RENDER_TARGET:
-            case RESOURCE_TYPE_DEPTH_STENCIL:
-            case RESOURCE_TYPE_VERTEX_BUFFER:
-                break;
-
-        }
+        ++interface_info.binding_count;
+        binding_idx += descriptor->count;
     }
 
     for (i = 0; i < runner->r.sampler_count; ++i)
@@ -481,13 +456,19 @@ static bool compile_shader(struct metal_runner *runner, enum shader_type type, s
         binding->shader_visibility = VKD3D_SHADER_VISIBILITY_ALL;
         binding->flags = 0;
         binding->binding.set = 0;
-        binding->binding.binding = interface_info.binding_count;
+        binding->binding.binding = binding_idx;
         binding->binding.count = 1;
+
         ++interface_info.binding_count;
+        ++binding_idx;
     }
 
     interface_info.bindings = bindings;
-    interface_info.next = &runner->signatures[type];
+    interface_info.next = &runner->thread_group_size;
+
+    runner->thread_group_size.type = VKD3D_SHADER_STRUCTURE_TYPE_SCAN_THREAD_GROUP_SIZE_INFO;
+    runner->thread_group_size.next = &runner->signatures[type];
+
     runner->signatures[type].type = VKD3D_SHADER_STRUCTURE_TYPE_SCAN_SIGNATURE_INFO;
     runner->signatures[type].next = NULL;
 
@@ -521,15 +502,43 @@ static id<MTLFunction> compile_stage(struct metal_runner *runner, enum shader_ty
     return [function autorelease];
 }
 
-static bool encode_argument_buffer(struct metal_runner *runner,
-        id<MTLRenderCommandEncoder> command_encoder, id<MTLSamplerState> *samplers)
+static struct metal_resource *find_resource(struct metal_runner *runner,
+        const struct descriptor_mapping *descriptor, unsigned int offset)
+{
+    enum resource_type resource_type;
+    struct resource *resource;
+
+    switch (descriptor->type)
+    {
+        case VKD3D_SHADER_DESCRIPTOR_TYPE_SRV:
+            resource_type = RESOURCE_TYPE_TEXTURE;
+            break;
+
+        case VKD3D_SHADER_DESCRIPTOR_TYPE_UAV:
+            resource_type = RESOURCE_TYPE_UAV;
+            break;
+
+        default:
+            fatal_error("Unhandled descriptor type %#x.\n", descriptor->type);
+    }
+
+    resource = shader_runner_get_resource(&runner->r, resource_type, descriptor->target_idx + offset);
+
+    if (!resource)
+        return NULL;
+
+    return metal_resource(resource);
+}
+
+static bool encode_argument_buffer(struct metal_runner *runner, id<MTLCommandEncoder> command_encoder,
+        void (*use_resource)(id<MTLCommandEncoder> encoder, id<MTLResource> resource, MTLResourceUsage usage),
+        const id<MTLSamplerState> *samplers, id<MTLBuffer> *argument_buffer)
 {
     NSMutableArray<MTLArgumentDescriptor *> *argument_descriptors;
     id<MTLDevice> device = runner->device;
     MTLArgumentDescriptor *arg_desc;
     id<MTLArgumentEncoder> encoder;
-    id<MTLBuffer> argument_buffer;
-    unsigned int i, index = 0;
+    unsigned int i, j, index = 0;
 
     argument_descriptors = [[[NSMutableArray alloc] init] autorelease];
 
@@ -542,34 +551,21 @@ static bool encode_argument_buffer(struct metal_runner *runner,
         [argument_descriptors addObject:arg_desc];
     }
 
-    for (i = 0; i < runner->r.resource_count; ++i)
+    for (i = 0; i < runner->r.descriptor_count; ++i)
     {
-        struct metal_resource *resource = metal_resource(runner->r.resources[i]);
+        const struct descriptor_mapping *descriptor = &runner->r.descriptors[i];
 
-        switch (resource->r.desc.type)
+        for (j = 0; j < descriptor->count; ++j)
         {
-            case RESOURCE_TYPE_TEXTURE:
-                arg_desc = [MTLArgumentDescriptor argumentDescriptor];
-                arg_desc.dataType = MTLDataTypeTexture;
-                arg_desc.index = [argument_descriptors count];
-                arg_desc.access = MTLBindingAccessReadOnly;
-                arg_desc.textureType = [resource->texture textureType];
-                [argument_descriptors addObject:arg_desc];
-                break;
+            const struct metal_resource *resource = find_resource(runner, descriptor, j);
 
-            case RESOURCE_TYPE_UAV:
-                arg_desc = [MTLArgumentDescriptor argumentDescriptor];
-                arg_desc.dataType = MTLDataTypeTexture;
-                arg_desc.index = [argument_descriptors count];
-                arg_desc.access = MTLBindingAccessReadWrite;
-                arg_desc.textureType = [resource->texture textureType];
-                [argument_descriptors addObject:arg_desc];
-                break;
-
-            case RESOURCE_TYPE_RENDER_TARGET:
-            case RESOURCE_TYPE_DEPTH_STENCIL:
-            case RESOURCE_TYPE_VERTEX_BUFFER:
-                break;
+            arg_desc = [MTLArgumentDescriptor argumentDescriptor];
+            arg_desc.dataType = MTLDataTypeTexture;
+            arg_desc.index = [argument_descriptors count];
+            arg_desc.access = descriptor->type == VKD3D_SHADER_DESCRIPTOR_TYPE_SRV ?
+                    MTLBindingAccessReadOnly : MTLBindingAccessReadWrite;
+            arg_desc.textureType = resource ? [resource->texture textureType] : MTLTextureType2D;
+            [argument_descriptors addObject:arg_desc];
         }
     }
 
@@ -583,12 +579,15 @@ static bool encode_argument_buffer(struct metal_runner *runner,
     }
 
     if (![argument_descriptors count])
+    {
+        *argument_buffer = nil;
         return true;
+    }
 
     encoder = [[device newArgumentEncoderWithArguments:argument_descriptors] autorelease];
-    argument_buffer = [[device newBufferWithLength:encoder.encodedLength
-            options:DEFAULT_BUFFER_RESOURCE_OPTIONS | MTLResourceStorageModeManaged] autorelease];
-    [encoder setArgumentBuffer:argument_buffer offset:0];
+    *argument_buffer = [device newBufferWithLength:encoder.encodedLength
+            options:DEFAULT_BUFFER_RESOURCE_OPTIONS | MTLResourceStorageModeManaged];
+    [encoder setArgumentBuffer:*argument_buffer offset:0];
 
     if (runner->r.uniform_count)
     {
@@ -598,54 +597,109 @@ static bool encode_argument_buffer(struct metal_runner *runner,
                 length:runner->r.uniform_count * sizeof(*runner->r.uniforms)
                 options:DEFAULT_BUFFER_RESOURCE_OPTIONS | MTLResourceStorageModeManaged] autorelease];
         [encoder setBuffer:cb offset:0 atIndex:index++];
-        [command_encoder useResource:cb
-                usage:MTLResourceUsageRead
-                stages:MTLRenderStageVertex | MTLRenderStageFragment];
+        use_resource(command_encoder, cb, MTLResourceUsageRead);
     }
 
-    for (i = 0; i < runner->r.resource_count; ++i)
+    for (i = 0; i < runner->r.descriptor_count; ++i)
     {
-        struct metal_resource *resource = metal_resource(runner->r.resources[i]);
+        const struct descriptor_mapping *descriptor = &runner->r.descriptors[i];
+        MTLResourceUsage usage;
 
-        switch (resource->r.desc.type)
+        usage = descriptor->type == VKD3D_SHADER_DESCRIPTOR_TYPE_SRV ?
+                MTLResourceUsageRead : MTLResourceUsageRead | MTLResourceUsageWrite;
+
+        for (j = 0; j < descriptor->count; ++j)
         {
-            case RESOURCE_TYPE_TEXTURE:
-                [encoder setTexture:resource->texture atIndex:index++];
-                [command_encoder useResource:resource->texture
-                        usage:MTLResourceUsageRead
-                        stages:MTLRenderStageVertex | MTLRenderStageFragment];
-                break;
+            const struct metal_resource *resource = find_resource(runner, descriptor, j);
 
-            case RESOURCE_TYPE_UAV:
-                [encoder setTexture:resource->texture atIndex:index++];
-                [command_encoder useResource:resource->texture
-                        usage:MTLResourceUsageRead | MTLResourceUsageWrite
-                        stages:MTLRenderStageVertex | MTLRenderStageFragment];
-                break;
+            if (resource)
+            {
+                [encoder setTexture:resource->texture atIndex:index];
+                use_resource(command_encoder, resource->texture, usage);
+            }
 
-            case RESOURCE_TYPE_RENDER_TARGET:
-            case RESOURCE_TYPE_DEPTH_STENCIL:
-            case RESOURCE_TYPE_VERTEX_BUFFER:
-                break;
+            ++index;
         }
     }
 
-    for (i = 0; i < runner->r.sampler_count; ++i)
+    if (samplers)
     {
-        [encoder setSamplerState:samplers[i] atIndex:index++];
+        for (i = 0; i < runner->r.sampler_count; ++i)
+        {
+            [encoder setSamplerState:samplers[i] atIndex:index++];
+        }
     }
 
-    [argument_buffer didModifyRange:NSMakeRange(0, encoder.encodedLength)];
-
-    [command_encoder setVertexBuffer:argument_buffer offset:0 atIndex:0];
-    [command_encoder setFragmentBuffer:argument_buffer offset:0 atIndex:0];
+    [*argument_buffer didModifyRange:NSMakeRange(0, encoder.encodedLength)];
 
     return true;
 }
 
+static void use_compute_resource(id<MTLCommandEncoder> encoder, id<MTLResource> resource, MTLResourceUsage usage)
+{
+    [(id<MTLComputeCommandEncoder>)encoder useResource:resource usage:usage];
+}
+
 static bool metal_runner_dispatch(struct shader_runner *r, unsigned int x, unsigned int y, unsigned int z)
 {
-    return false;
+    struct metal_runner *runner = metal_runner(r);
+    id<MTLDevice> device = runner->device;
+    id<MTLComputeCommandEncoder> encoder;
+    id<MTLCommandBuffer> command_buffer;
+    id<MTLComputePipelineState> pso;
+    id<MTLBuffer> argument_buffer;
+    id<MTLFunction> cs;
+    NSError *err;
+
+    if (!(cs = compile_stage(runner, SHADER_TYPE_CS)))
+    {
+        trace("Failed to compile kernel function.\n");
+        return false;
+    }
+
+    pso = [device newComputePipelineStateWithFunction:cs error:&err];
+    [cs release];
+    if (!pso)
+    {
+        trace("Failed to compile pipeline state.\n");
+        if (err)
+        {
+            trace_messages([err.localizedDescription UTF8String]);
+            [err release];
+        }
+        return false;
+    }
+
+    command_buffer = [runner->queue commandBuffer];
+
+    encoder = [command_buffer computeCommandEncoder];
+
+    if (!encode_argument_buffer(runner, encoder, use_compute_resource, NULL, &argument_buffer))
+    {
+        [encoder endEncoding];
+        [encoder release];
+        [command_buffer release];
+        [pso release];
+        return false;
+    }
+    [encoder setBuffer:argument_buffer offset:0 atIndex:0];
+
+    [encoder setComputePipelineState:pso];
+    [encoder dispatchThreadgroups:MTLSizeMake(x, y, z)
+            threadsPerThreadgroup:MTLSizeMake(runner->thread_group_size.x,
+                    runner->thread_group_size.y, runner->thread_group_size.z)];
+
+    [encoder endEncoding];
+    [encoder release];
+
+    [command_buffer commit];
+    [command_buffer waitUntilCompleted];
+    [command_buffer release];
+
+    [argument_buffer release];
+    [pso release];
+
+    return true;
 }
 
 static void metal_runner_clear(struct shader_runner *r, struct resource *res, const struct vec4 *clear_value)
@@ -688,6 +742,13 @@ static void metal_runner_clear(struct shader_runner *r, struct resource *res, co
     }
 }
 
+static void use_graphics_resource(id<MTLCommandEncoder> encoder, id<MTLResource> resource, MTLResourceUsage usage)
+{
+    [(id<MTLRenderCommandEncoder>)encoder useResource:resource
+                                                usage:usage
+                                               stages:MTLRenderStageVertex | MTLRenderStageFragment];
+}
+
 static bool metal_runner_draw(struct shader_runner *r, D3D_PRIMITIVE_TOPOLOGY topology,
         unsigned int vertex_count, unsigned int instance_count)
 {
@@ -703,6 +764,7 @@ static bool metal_runner_draw(struct shader_runner *r, D3D_PRIMITIVE_TOPOLOGY to
     size_t attribute_offsets[32], stride;
     id<MTLRenderCommandEncoder> encoder;
     id<MTLCommandBuffer> command_buffer;
+    id<MTLBuffer> argument_buffer = nil;
     MTLDepthStencilDescriptor *ds_desc;
     MTLRenderPassDescriptor *pass_desc;
     MTLSamplerDescriptor *sampler_desc;
@@ -746,9 +808,9 @@ static bool metal_runner_draw(struct shader_runner *r, D3D_PRIMITIVE_TOPOLOGY to
             sampler_desc.rAddressMode = get_metal_address_mode(sampler->w_address);
             sampler_desc.magFilter = (sampler->filter & 0x4)
                     ? MTLSamplerMinMagFilterLinear : MTLSamplerMinMagFilterNearest;
-            sampler_desc.minFilter = (sampler->filter & 0x1)
+            sampler_desc.minFilter = (sampler->filter & 0x10)
                     ? MTLSamplerMinMagFilterLinear : MTLSamplerMinMagFilterNearest;
-            sampler_desc.mipFilter = (sampler->filter & 0x10)
+            sampler_desc.mipFilter = (sampler->filter & 0x1)
                     ? MTLSamplerMipFilterLinear : MTLSamplerMipFilterNearest;
             sampler_desc.compareFunction = sampler->func
                     ? get_metal_compare_function(sampler->func) : MTLCompareFunctionNever;
@@ -826,12 +888,15 @@ static bool metal_runner_draw(struct shader_runner *r, D3D_PRIMITIVE_TOPOLOGY to
         command_buffer = [runner->queue commandBuffer];
         encoder = [command_buffer renderCommandEncoderWithDescriptor:pass_desc];
 
-        if (!encode_argument_buffer(runner, encoder, samplers))
+        if (!encode_argument_buffer(runner, encoder, use_graphics_resource, samplers, &argument_buffer))
         {
             [encoder endEncoding];
             ret = false;
             goto done;
         }
+        [encoder setVertexBuffer:argument_buffer offset:0 atIndex:0];
+        [encoder setFragmentBuffer:argument_buffer offset:0 atIndex:0];
+        [argument_buffer autorelease];
 
         if (runner->r.input_element_count > 32)
             fatal_error("Unsupported input element count %zu.\n", runner->r.input_element_count);
@@ -1114,8 +1179,35 @@ static bool metal_runner_init(struct metal_runner *runner)
     }
 
     runner->caps.runner = "Metal";
-    runner->caps.tags[0] = "msl";
-    runner->caps.tag_count = 1;
+    runner->caps.shader_caps[SHADER_CAP_DESCRIPTORS] = true;
+
+    runner->caps.tags[SHADER_RUNNER_TAG_MSL].present = true;
+
+    if (runner->device.readWriteTextureSupport >= MTLReadWriteTextureTier1)
+    {
+        runner->caps.format_caps[DXGI_FORMAT_R32_UINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R32_SINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R32_FLOAT] |= FORMAT_CAP_UAV_LOAD;
+    }
+
+    if (runner->device.readWriteTextureSupport >= MTLReadWriteTextureTier2)
+    {
+        runner->caps.format_caps[DXGI_FORMAT_R32G32B32A32_UINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R32G32B32A32_SINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R32G32B32A32_FLOAT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R16G16B16A16_UINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R16G16B16A16_SINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R16G16B16A16_FLOAT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R8G8B8A8_UNORM] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R8G8B8A8_UINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R8G8B8A8_SINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R16_UINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R16_SINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R16_FLOAT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R8_UNORM] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R8_UINT] |= FORMAT_CAP_UAV_LOAD;
+        runner->caps.format_caps[DXGI_FORMAT_R8_SINT] |= FORMAT_CAP_UAV_LOAD;
+    }
 
     return true;
 }
