@@ -354,6 +354,7 @@ static bool compile_hlsl_and_scan(struct vulkan_shader_runner *runner, enum shad
 static bool compile_d3d_code(struct vulkan_shader_runner *runner,
         enum shader_type type, struct vkd3d_shader_code *spirv)
 {
+    struct vkd3d_shader_d3dbc_source_info d3dbc_source_info = {.type = VKD3D_SHADER_STRUCTURE_TYPE_D3DBC_SOURCE_INFO};
     struct vkd3d_shader_varying_map_info varying_map_info = {.type = VKD3D_SHADER_STRUCTURE_TYPE_VARYING_MAP_INFO};
     struct vkd3d_shader_spirv_target_info spirv_info = {.type = VKD3D_SHADER_STRUCTURE_TYPE_SPIRV_TARGET_INFO};
     struct vkd3d_shader_interface_info interface_info = {.type = VKD3D_SHADER_STRUCTURE_TYPE_INTERFACE_INFO};
@@ -365,13 +366,13 @@ static bool compile_d3d_code(struct vulkan_shader_runner *runner,
     struct vkd3d_shader_varying_map varying_map[12];
     struct vkd3d_shader_resource_binding *binding;
     struct vkd3d_shader_compile_option options[2];
-    struct vkd3d_shader_parameter1 parameters[22];
+    struct vkd3d_shader_parameter1 parameters[40];
     unsigned int i;
     char *messages;
     int ret;
 
     options[0].name = VKD3D_SHADER_COMPILE_OPTION_API_VERSION;
-    options[0].value = VKD3D_SHADER_API_VERSION_1_17;
+    options[0].value = VKD3D_SHADER_API_VERSION_CURRENT;
     options[1].name = VKD3D_SHADER_COMPILE_OPTION_FEATURE;
     options[1].value = shader_runner_caps_get_feature_flags(&runner->caps);
 
@@ -398,6 +399,28 @@ static bool compile_d3d_code(struct vulkan_shader_runner *runner,
         ok(varying_map_info.varying_count <= ARRAY_SIZE(varying_map),
                 "Got unexpected count %u.\n", varying_map_info.varying_count);
     }
+    else if (type == SHADER_TYPE_PS && info.source_type == VKD3D_SHADER_SOURCE_D3D_BYTECODE)
+    {
+        info.next = &d3dbc_source_info;
+
+        d3dbc_source_info.next = &spirv_info;
+
+        for (i = 0; i < runner->r.resource_count; ++i)
+        {
+            const struct resource *resource = runner->r.resources[i];
+
+            if (resource->desc.type != RESOURCE_TYPE_TEXTURE
+                    || resource->desc.slot > ARRAY_SIZE(d3dbc_source_info.texture_dimensions))
+                continue;
+
+            if (resource->desc.dimension == RESOURCE_DIMENSION_3D)
+                d3dbc_source_info.texture_dimensions[resource->desc.slot] = VKD3D_SHADER_RESOURCE_TEXTURE_3D;
+            else if (resource->desc.dimension == RESOURCE_DIMENSION_CUBE)
+                d3dbc_source_info.texture_dimensions[resource->desc.slot] = VKD3D_SHADER_RESOURCE_TEXTURE_CUBE;
+            else
+                d3dbc_source_info.texture_dimensions[resource->desc.slot] = VKD3D_SHADER_RESOURCE_TEXTURE_2D;
+        }
+    }
 
     spirv_info.next = &interface_info;
     spirv_info.environment = VKD3D_SHADER_SPIRV_ENVIRONMENT_VULKAN_1_0;
@@ -568,6 +591,24 @@ static bool compile_d3d_code(struct vulkan_shader_runner *runner,
     parameters[21].data_type = VKD3D_SHADER_PARAMETER_DATA_TYPE_UINT32;
     parameters[21].u.immediate_constant.u.u32 = get_fog_source(&runner->r);
 
+    for (i = 0; i < 6; ++i)
+    {
+        parameters[22 + i].name = VKD3D_SHADER_PARAMETER_NAME_BUMP_MATRIX_0 + i;
+        parameters[22 + i].type = VKD3D_SHADER_PARAMETER_TYPE_IMMEDIATE_CONSTANT;
+        parameters[22 + i].data_type = VKD3D_SHADER_PARAMETER_DATA_TYPE_FLOAT32_VEC4;
+        memcpy(parameters[22 + i].u.immediate_constant.u.f32_vec4, runner->r.bump[i].matrix, sizeof(struct vec4));
+
+        parameters[28 + i].name = VKD3D_SHADER_PARAMETER_NAME_BUMP_LUMINANCE_SCALE_0 + i;
+        parameters[28 + i].type = VKD3D_SHADER_PARAMETER_TYPE_IMMEDIATE_CONSTANT;
+        parameters[28 + i].data_type = VKD3D_SHADER_PARAMETER_DATA_TYPE_FLOAT32;
+        parameters[28 + i].u.immediate_constant.u.f32 = runner->r.bump[i].luminance_scale;
+
+        parameters[34 + i].name = VKD3D_SHADER_PARAMETER_NAME_BUMP_LUMINANCE_OFFSET_0 + i;
+        parameters[34 + i].type = VKD3D_SHADER_PARAMETER_TYPE_IMMEDIATE_CONSTANT;
+        parameters[34 + i].data_type = VKD3D_SHADER_PARAMETER_DATA_TYPE_FLOAT32;
+        parameters[34 + i].u.immediate_constant.u.f32 = runner->r.bump[i].luminance_offset;
+    }
+
     parameter_info.parameter_count = ARRAY_SIZE(parameters);
     parameter_info.parameters = parameters;
 
@@ -1003,8 +1044,8 @@ static VkDescriptorSetLayout create_descriptor_set_layout(struct vulkan_shader_r
         const struct sampler *sampler = &runner->r.samplers[i];
 
         sampler_desc.magFilter = (sampler->filter & 0x4) ? VK_FILTER_LINEAR : VK_FILTER_NEAREST;
-        sampler_desc.minFilter = (sampler->filter & 0x1) ? VK_FILTER_LINEAR : VK_FILTER_NEAREST;
-        sampler_desc.mipmapMode = (sampler->filter & 0x10) ? VK_SAMPLER_MIPMAP_MODE_LINEAR : VK_SAMPLER_MIPMAP_MODE_NEAREST;
+        sampler_desc.minFilter = (sampler->filter & 0x10) ? VK_FILTER_LINEAR : VK_FILTER_NEAREST;
+        sampler_desc.mipmapMode = (sampler->filter & 0x1) ? VK_SAMPLER_MIPMAP_MODE_LINEAR : VK_SAMPLER_MIPMAP_MODE_NEAREST;
         sampler_desc.addressModeU = vk_address_mode_from_d3d12(sampler->u_address);
         sampler_desc.addressModeV = vk_address_mode_from_d3d12(sampler->v_address);
         sampler_desc.addressModeW = vk_address_mode_from_d3d12(sampler->w_address);
@@ -1779,21 +1820,25 @@ static bool init_vulkan_runner(struct vulkan_shader_runner *runner)
     get_physical_device_info(runner, &device_info);
     ret_features = &device_info.features2.features;
 
-    runner->caps.tag_count = 0;
-    runner->caps.tags[runner->caps.tag_count++] = "vulkan";
-    if (device_info.driver_properties.driverID == VK_DRIVER_ID_MOLTENVK)
-    {
-        runner->caps.tags[runner->caps.tag_count++] = "mvk";
-    }
-    else
-    {
-        if (device_info.driver_properties.driverID == VK_DRIVER_ID_MESA_LLVMPIPE)
-            runner->caps.tags[runner->caps.tag_count++] = "llvmpipe";
-        if (is_mesa_vulkan_driver(&device_info.driver_properties)
-                && !is_vulkan_driver_version_ge(&device_info.properties2.properties,
-                        &device_info.driver_properties, 23, 3, 0))
-            runner->caps.tags[runner->caps.tag_count++] = "mesa<23.3";
-    }
+    runner->caps.tags[SHADER_RUNNER_TAG_VULKAN].present = true;
+
+    runner->caps.tags[SHADER_RUNNER_TAG_MVK].present = get_mvk_driver_version(&device_info.properties2.properties,
+            &device_info.driver_properties, &runner->caps.tags[SHADER_RUNNER_TAG_MVK].major,
+            &runner->caps.tags[SHADER_RUNNER_TAG_MVK].minor, &runner->caps.tags[SHADER_RUNNER_TAG_MVK].patch);
+
+    runner->caps.tags[SHADER_RUNNER_TAG_MESA].present = get_mesa_driver_version(&device_info.properties2.properties,
+            &device_info.driver_properties, &runner->caps.tags[SHADER_RUNNER_TAG_MESA].major,
+            &runner->caps.tags[SHADER_RUNNER_TAG_MESA].minor, &runner->caps.tags[SHADER_RUNNER_TAG_MESA].patch);
+
+    runner->caps.tags[SHADER_RUNNER_TAG_LLVM].present = get_llvm_driver_version(
+            &device_info.properties2.properties, &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].major,
+            &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].minor, &runner->caps.tags[SHADER_RUNNER_TAG_LLVM].patch);
+
+    if (device_info.driver_properties.driverID == VK_DRIVER_ID_MESA_LLVMPIPE)
+        runner->caps.tags[SHADER_RUNNER_TAG_LLVMPIPE].present = true;
+
+    if (device_info.driver_properties.driverID == VK_DRIVER_ID_INTEL_OPEN_SOURCE_MESA)
+        runner->caps.tags[SHADER_RUNNER_TAG_INTEL].present = true;
 
     runner->caps.shader_caps[SHADER_CAP_CLIP_PLANES] = true;
     runner->caps.shader_caps[SHADER_CAP_FOG] = true;
@@ -1835,6 +1880,12 @@ static bool init_vulkan_runner(struct vulkan_shader_runner *runner)
         runner->caps.shader_caps[SHADER_CAP_GEOMETRY_SHADER] = true;
     }
 
+    if (ret_features->shaderCullDistance)
+    {
+        features.shaderCullDistance = VK_TRUE;
+        runner->caps.shader_caps[SHADER_CAP_CULL_DISTANCE] = true;
+    }
+
     if (ret_features->shaderFloat64)
     {
         features.shaderFloat64 = VK_TRUE;
