@@ -42,6 +42,7 @@
 #include <stdbool.h>
 #include <stdint.h>
 #include <stdlib.h>
+#include <math.h>
 #include "vkd3d_test.h"
 #include "vkd3d_d3d12.h"
 #include "vkd3d_d3dcompiler.h"
@@ -165,8 +166,13 @@ static bool compare_float(float f, float g, unsigned int ulps)
 
     u.f = f;
     x = u.i;
+    if (isnan(f))
+        x = (x & 0xff800000) | 0x400000;
+
     u.f = g;
     y = u.i;
+    if (isnan(g))
+        y = (y & 0xff800000) | 0x400000;
 
     if (x < 0)
         x = INT_MIN - x;
@@ -187,8 +193,13 @@ static bool compare_double(double f, double g, unsigned int ulps)
 
     u.f = f;
     x = u.i;
+    if (isnan(f))
+        x = (x & 0xfff0000000000000ll) | 0x8000000000000ll;
+
     u.f = g;
     y = u.i;
+    if (isnan(g))
+        y = (y & 0xfff0000000000000ll) | 0x8000000000000ll;
 
     if (x < 0)
         x = INT64_MIN - x;
@@ -580,8 +591,16 @@ static inline HRESULT vkd3d_shader_code_from_dxc_blob(IDxcBlob *blob, struct vkd
     return S_OK;
 }
 
-static inline HRESULT dxc_compile(void *dxc_compiler, const WCHAR *profile,
-        unsigned int compile_options, bool enable_16bit_types, const char *hlsl, struct vkd3d_shader_code *blob_out)
+enum denorm_mode
+{
+    DENORM_NOT_SPECIFIED = 0,
+    DENORM_PRESERVE,
+    DENORM_FTZ,
+    DENORM_ANY,
+};
+
+static inline HRESULT dxc_compile(void *dxc_compiler, const WCHAR *profile, unsigned int compile_options,
+        bool enable_16bit_types, enum denorm_mode denorm_mode, const char *hlsl, struct vkd3d_shader_code *blob_out)
 {
     DxcBuffer src_buf = {hlsl, strlen(hlsl), 65001};
     IDxcCompiler3 *compiler = dxc_compiler;
@@ -604,11 +623,12 @@ static inline HRESULT dxc_compile(void *dxc_compiler, const WCHAR *profile,
         NULL,
         NULL,
         NULL,
+        NULL,
     };
 
     memset(blob_out, 0, sizeof(*blob_out));
 
-    arg_count = ARRAY_SIZE(args) - 4;
+    arg_count = ARRAY_SIZE(args) - 5;
     if (compile_options & D3DCOMPILE_PACK_MATRIX_ROW_MAJOR)
         args[arg_count++] = L"/Zpr";
     if (compile_options & D3DCOMPILE_PACK_MATRIX_COLUMN_MAJOR)
@@ -618,6 +638,24 @@ static inline HRESULT dxc_compile(void *dxc_compiler, const WCHAR *profile,
     if (enable_16bit_types)
         args[arg_count++] = L"/enable-16bit-types";
 
+    switch (denorm_mode)
+    {
+        case DENORM_NOT_SPECIFIED:
+            break;
+
+        case DENORM_PRESERVE:
+            args[arg_count++] = L"/denorm preserve";
+            break;
+
+        case DENORM_FTZ:
+            args[arg_count++] = L"/denorm ftz";
+            break;
+
+        case DENORM_ANY:
+            args[arg_count++] = L"/denorm any";
+            break;
+    }
+
     if (FAILED(hr = IDxcCompiler3_Compile(compiler, &src_buf, args,
             arg_count, NULL, &IID_IDxcResult, (void **)&result)))
         return hr;
@@ -691,4 +729,29 @@ static inline IDxcCompiler3 *dxcompiler_create(void)
 }
 #endif
 
+static inline int compare_versions(uint32_t major1, uint32_t minor1, uint32_t patch1,
+        uint32_t major2, uint32_t minor2, uint32_t patch2)
+{
+    int ret;
+
+    if ((ret = vkd3d_u32_compare(major1, major2)))
+        return ret;
+    if ((ret = vkd3d_u32_compare(minor1, minor2)))
+        return ret;
+    return vkd3d_u32_compare(patch1, patch2);
+}
+
+static inline bool find_version_string(const char *str, const char *tag,
+        uint32_t *major, uint32_t *minor, uint32_t *patch)
+{
+    const char *ptr;
+
+    if (!(ptr = strstr(str, tag)))
+        return false;
+
+    ptr += strlen(tag);
+
+    return 3 == sscanf(ptr, "%u.%u.%u", major, minor, patch);
+}
+
 #endif
